#!/usr/bin/env python3
# Author: Simeon Reusch (simeon.reusch@desy.de)
# License: BSD-3-Clause

import backoff  # type: ignore
import requests

from betternot import credentials

FRITZ_TOKEN = credentials.get_credentials(service="FRITZ", token=True)["token"]
BASE_URL = "https://fritz.science/api"


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)
def api(
    method: str, url: str, data: dict | None = None, stream: bool = False
) -> requests.Response:
    """
    Basic API request method
    """
    headers = {"Authorization": f"token {FRITZ_TOKEN}"}

    endpoint = BASE_URL + url
    response = requests.request(
        method=method, url=endpoint, json=data, headers=headers, stream=stream
    )

    if response.status_code != 200:
        raise requests.exceptions.RequestException

    return response


def radec(ztf_id: str):
    """
    Get RA and Dec of a source, specified by its ZTF-ID
    """
    response = api(method="get", url=f"/sources/{ztf_id}")

    res = response.json()
    ra = res["data"].get("ra")
    dec = res["data"].get("dec")

    return (ra, dec)


def latest_photometry(ztf_id: str):
    """
    Retrieve the photometry of a source, specified by its ZTF-ID
    """
    response = api(method="get", url=f"/sources/{ztf_id}/photometry")

    phot_all = response.json().get("data")
    # latest = photdata[-1]
    # mag = latest[""]
    phot_dets = []
    for entry in phot_all:
        if entry["mag"] != None:
            phot_dets.append(entry)

    latest = phot_dets[-1]
    mag = latest["mag"]
    mjd = latest["mjd"]
    band = latest["filter"]

    return (mag, mjd, band)



#!/usr/bin/env python3
# Author: Simeon Reusch (simeon.reusch@desy.de)
# License: BSD-3-Clause

from pathlib import Path

import yaml  # type: ignore

basedir = Path(__file__).parents[1]


def get_object_dir(ztf_id: str) -> Path:
    directory = basedir / ztf_id
    directory.mkdir(parents=True, exist_ok=True)

    return directory


def get_date_dir(date: str) -> Path:
    directory = basedir / date
    directory.mkdir(parents=True, exist_ok=True)

    return directory


def load_config() -> dict:
    """
    Load the config (contains e.g. standard stars)
    """
    current_dir = Path(__file__)
    config_dir = current_dir.parents[1]
    config_file = config_dir / "config.yaml"

    with open(config_file, "r") as stream:
        config = yaml.safe_load(stream)

    return config



#!/usr/bin/env python3

import datetime
import json
import logging
import os
import sys
import time
from collections import OrderedDict
from pathlib import Path

import requests
import yaml
from astropy.io import ascii  # type: ignore

from betternot import credentials
from betternot.fritz import radec

TNS_TOKEN = credentials.get_credentials(service="TNS", token=True)["token"]
TNS_BOT_ID = "115364"
TNS_BOT_NAME = "ZTF_DESY"
WISEREP_TOKEN = credentials.get_credentials(service="WISEREP", token=True)["token"]
WISEREP_BOT_ID = "1234"
WISEREP_BOT_NAME = "OKC_ZTF"


class Wiserep:
    """
    Upload a spectrum to WISeREP
    """

    def __init__(
        self,
        ztf_id: str,
        spec_path: Path | str,
        quality: str = "medium",
        sandbox: bool = True,
        tns_name: str | None = None,
    ):
        self.logger = logging.getLogger()
        self.ztf_id = ztf_id
        self.spec_path = Path(spec_path)
        self.quality = quality

        if sandbox:
            self.wiserep_endpoint = "https://sandbox.wiserep.org/api"
        else:
            self.wiserep_endpoint = "https://www.wiserep.org/api"

        self.ra, self.dec = radec(self.ztf_id)

        if tns_name is None:
            self.tns_name = self.query_tns()
        else:
            self.tns_name = tns_name

        if self.tns_name is not None:
            server_filename = self.upload_files([self.spec_path])[0]
            if server_filename is not None:
                self.read_spectrum(server_filename=server_filename)
                self.generate_report()
                res = self.send_metadata()
                self.res = res

    def query_tns(self) -> str | None:
        """
        Check if the object is known on TNS (so we can use the ID on WISeREP, I have not figured out how to do a WISeREP cone search.)
        """
        queryurl_tns = "https://www.wis-tns.org/api/get/search"

        tns_marker = (
            'tns_marker{"tns_id": "'
            + str(TNS_BOT_ID)
            + '", "type": "bot", "name": "'
            + TNS_BOT_NAME
            + '"}'
        )
        headers = {"User-Agent": tns_marker}

        get_obj = {
            "ra": self.ra,
            "dec": self.dec,
            "radius": 3,
            "units": "arcsec",
        }

        json_string = json.dumps(get_obj)

        # I have no idea why the token is not in the header
        payload = {"api_key": TNS_TOKEN, "data": json_string}

        print(queryurl_tns)
        print(headers)
        print(payload)

        response = requests.post(queryurl_tns, headers=headers, data=payload)

        if response.status_code != 200:
            self.logger.warn(
                f"You got status code {response.status_code}. Something went wrong"
            )
            tns_name = None

        res_json = response.json()
        reply = res_json["data"]["reply"]

        if len(reply) > 0:
            tns_name = reply[0]["objname"]
            self.logger.info(f"Found match on TNS: {tns_name}")
        else:
            tns_name = None

            self.logger.info("Found no match on TNS.")

        return tns_name

    def read_spectrum(self, server_filename: str | None = None):
        """
        Open the spectrum ascii file and extract metadata
        """
        data = ascii.read(self.spec_path, names=("WAVE", "FLUX", "FLUX_ERR"))
        meta = data.meta["comments"]
        metadict = {}

        ncombine = 1
        exptime = 0

        for entry in meta:
            keyval = entry.split("=")
            if len(keyval) > 1:
                key = keyval[0]
                val = keyval[1]
                if key == "HOME_OBSERVER" or key == "OBSERVER":
                    observer = val
                    metadict.update({"observer": val})
                elif key == "REDUCER":
                    metadict.update({"reducer": val})
                elif key == "EXPTIME":
                    exptime = int(float(val))
                elif key == "DATE-OBS":
                    metadict.update({"obsdate": val.replace("T", " ")})
                elif key == "NCOMBINE":
                    ncombine = int(val)

        inttime = ncombine * exptime
        metadict.update({"exptime": inttime})

        if server_filename is not None:
            metadict.update({"ascii_file": server_filename})
        else:
            metadict.update({"ascii_file": str(self.spec_path)})

        # Let's assume the last modification time of the spectrum file is the reduction time
        timestamp = self.spec_path.stat().st_mtime
        reducedate = str(datetime.datetime.fromtimestamp(timestamp))
        metadict.update({"reduction_date": reducedate})

        self.metadata = metadict

    def generate_report(self, tns_name: str | None = None):
        """
        Open and fill the template spectrum report with the metadata of the spectrum
        """
        template_path = Path(__file__).parent.parent / "data" / "template.yaml"
        with open(template_path, "r") as stream:
            report = yaml.safe_load(stream)

        for key, val in self.metadata.items():
            report["objects"][0]["spectra"]["spectra_group"][0][key] = val

        if tns_name is None:
            tns_name = self.tns_name

        report["objects"][0]["iau_name"] = tns_name
        report["objects"][0]["ra"] = self.ra
        report["objects"][0]["decl"] = self.dec

        quality_levels = {"low": "1", "medium": "2", "high": "3"}
        report["objects"][0]["spectra"]["spectra_group"][0][
            "qualityid"
        ] = quality_levels[self.quality]

        self.report = report

    def upload_files(self, file_list: list[Path]) -> list:
        """
        Upload a file to WISErEP and check the response
        """
        self.logger.info(
            f"Uploading {' '.join(str(x) for x in file_list)} to the WISeREP"
        )
        url = self.wiserep_endpoint + "/file-upload"

        headers = {
            "User-Agent": 'tns_marker{"tns_id":'
            + str(WISEREP_BOT_ID)
            + ', "type":"bot",'
            ' "name":"' + WISEREP_BOT_NAME + '"}'
        }

        # api key data
        api_data = {"bot_api_key": WISEREP_TOKEN}

        # construct a dictionary of files and file handles
        files_data = {}
        for i, path in enumerate(file_list):
            key = "files[" + str(i) + "]"
            val = (str(path), open(path), "text/plain")
            files_data[key] = val

        response = requests.post(url, headers=headers, data=api_data, files=files_data)

        if response.status_code == 200:
            server_filenames = response.json()["data"]
            self.logger.info(
                f"Received and saved as {server_filenames} on the WISeREP server"
            )
            return server_filenames
        else:
            self.logger.warn(
                f"Something went wrong. Reponse code: {response.status_code}"
            )
            return [None]

    # function for sending json metadata
    def send_json_report(self, json_report: str):
        report_url = self.wiserep_endpoint + "/bulk-report"
        # headers
        headers = {
            "User-Agent": 'tns_marker{"tns_id":'
            + str(WISEREP_BOT_ID)
            + ', "type":"bot",'
            ' "name":"' + WISEREP_BOT_NAME + '"}'
        }

        payload = {"bot_api_key": WISEREP_TOKEN, "data": json_report}

        response = requests.post(report_url, headers=headers, data=payload)

        if response.status_code == 200:
            self.logger.info("Sent metadata to WISeREP")
            return response.json()
        else:
            self.logger.warn(
                f"Something went wrong. Reponse code: {response.status_code}"
            )
            return None

    def send_metadata(self, report: str | None = None):
        """
        Send the metadata for a spectrum to WISeREP
        """
        if report is None:
            report = self.report

        json_report = json.dumps(report)

        res = self.send_json_report(json_report=json_report)

        if res is not None:
            self.logger.debug(res)

        return res



#!/usr/bin/env python3
# Author: Most of the code is refactored code by Steve Schulze (steve.schulze@fysik.su.se)
# License: BSD-3-Clause

import datetime
import logging
import warnings

import astroplan as ap  # type: ignore
import astropy  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
from astroplan import moon as apmoon  # type: ignore
from astroplan.plots import plot_airmass, plot_altitude  # type: ignore
from astropy import units as u  # type: ignore
from astropy.coordinates import AltAz, EarthLocation, SkyCoord, get_body  # type: ignore
from astropy.time import Time  # type: ignore

from betternot.io import get_date_dir, load_config


class Observability:
    def __init__(self, ztf_ids, date: str | None = None, site: str = "not"):
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)

        self.ztf_ids = ztf_ids

        if date is None:
            self.date = datetime.date.today().strftime("%Y-%m-%d")
        else:
            self.date = date

        self.config = load_config()
        self.site = EarthLocation.of_site(self.config["sites"][site]["short"])
        self.target_dict: dict = {}
        self.logger.info(
            f"Getting observation data for {', '.join(ztf_ids)} for the {self.config['sites'][site]['pretty']}. Chosen date: {self.date}"
        )

    def get_info(self):
        from betternot.fritz import radec, latest_photometry

        for ztf_id in self.ztf_ids:
            if ztf_id not in self.target_dict.keys():
                ra, dec = radec(ztf_id)
                mag, mjd, band = latest_photometry(ztf_id)

                if ra is None:
                    self.logger.info(f"Source {ztf_id} not found on Fritz, skipping.")
                else:
                    self.target_dict.update(
                        {
                            ztf_id: {
                                "ra": ra,
                                "dec": dec,
                                "mag": mag,
                                "mjd": mjd,
                                "band": band,
                            }
                        }
                    )

    def print_info(self):
        """
        Print the most importan information to enter when preparing an OB
        """
        str_to_print = ""

        for ztf_id, info in self.target_dict.items():
            now = Time.now().mjd
            days_ago = now - info["mjd"]
            coords = SkyCoord(
                info["ra"],
                info["dec"],
                unit=(u.deg, u.deg),
            )
            coords_str = coords.to_string(style="hmsdms")
            ra, dec = coords_str.split(" ")
            ra = ra.replace("h", ":").replace("m", ":").replace("s", "")
            dec = dec.replace("d", ":").replace("m", ":").replace("s", "")
            str_to_print += "-------------------------------------------\n"
            str_to_print += f"{ztf_id}\n"
            str_to_print += f"ztf{ztf_id[3:]}\n"
            str_to_print += f"RA: {ra}\n"
            str_to_print += f"Dec: {dec}\n"
            str_to_print += f"{info['mag']:.2f} mag {days_ago:.0f} days ago in the {info['band']} filter\n"
            str_to_print += "-------------------------------------------\n"

        print(str_to_print)

        self.info = str_to_print

    def plot_standards(self):
        std_dict = self.config["standards"]
        self.create_plot(target_dict=std_dict, savename="standards")

    def plot_targets(self):
        self.get_info()

        self.create_plot(
            target_dict=self.target_dict, savename="targets", plot_moon=True
        )

    def create_plot(self, target_dict: dict, savename: str, plot_moon: bool = False):
        target_list = list(target_dict)
        label_size = 14

        plt.figure(figsize=(width := 9, width / 1.61))
        ax = plt.subplot(111)

        self.midnight_utc = Time(self.date, format="isot", scale="utc") + (1 * u.hour)
        delta_midnight = np.linspace(-12, 12, 1000) * u.hour

        frame_time = AltAz(
            obstime=self.midnight_utc + delta_midnight, location=self.site
        )

        for target in target_list:
            if isinstance(target_dict[target]["ra"], str):
                coords = SkyCoord(
                    target_dict[target]["ra"],
                    target_dict[target]["dec"],
                    unit=(u.hour, u.deg),
                )
            else:
                coords = SkyCoord(
                    target_dict[target]["ra"],
                    target_dict[target]["dec"],
                    unit=(u.deg, u.deg),
                )

            obj_altazs = coords.transform_to(frame_time)

            if plot_moon:
                moon_info = self.check_moon(coords=coords)
                label = f"{target} (moon dist: {moon_info['sep']:.0f}°)"
                # illumsymbol = self.get_moon_emoticon(moon_info=moon_info)
            else:
                label = target

            ax.plot(
                delta_midnight,
                obj_altazs.alt,
                label=label,
                lw=2,
            )

        sunaltazs = get_body("sun", self.midnight_utc + delta_midnight).transform_to(
            frame_time
        )

        if plot_moon:
            moonaltazs = get_body(
                "moon", self.midnight_utc + delta_midnight
            ).transform_to(frame_time)
            ax.plot(delta_midnight, moonaltazs.alt, lw=2, color="white")

        for sunheight, alpha in [(-0, 0.2), (-18, 1)]:
            ax.fill_between(
                x=delta_midnight.value,
                y1=(0 * u.deg).value,
                y2=(90 * u.deg).value,
                where=sunaltazs.alt.value < (sunheight * u.deg).value,
                color="navy",
                zorder=0,
                alpha=alpha,
            )

        # Plot an airmass scale
        ax2 = ax.secondary_yaxis(
            "right", functions=(self.altitude_to_airmass, self.airmass_to_altitude)
        )
        altitude_ticks = np.linspace(10, 90, 9)
        airmass_ticks = np.round(self.altitude_to_airmass(altitude_ticks), 2)
        ax2.set_yticks(airmass_ticks)
        ax2.set_ylabel("Airmass", fontsize=label_size)

        ax.axvline(x=0, color="white", lw=1)
        ax.axhline(90 - np.arccos(1 / 2.0) * 180 / np.pi, color="#C02F1D")
        ax.axhline(90 - np.arccos(1 / 3.0) * 180 / np.pi, color="#C02F1D", ls="--")

        xmin, xmax = -6, 8

        ax.set_xlim((xmin * u.hour).value, (xmax * u.hour).value)

        labels = [
            "{:.0f} h".format(x + 24) if x < 0 else "{:.0f} h".format(x)
            for x in (np.arange(xmax - xmin) + xmin)
        ]

        ax.set_xticks(
            ((np.arange(xmax - xmin) + xmin) * u.hour).value, labels=labels, rotation=45
        )

        ax.set_ylim((10 * u.deg).value, (90 * u.deg).value)

        ax.set_xlabel("Universal time", fontsize=label_size)
        ax.set_ylabel("Altitude", fontsize=label_size)

        title = f"{self.date} → {(Time(self.date) + (1 * u.day)).isot.split('T')[0]}"

        if plot_moon:
            # title += f" {illumsymbol} ({moon_info['illum']:.0f} %)"
            title += f"{moon_info['illum']:.0f} %"

            ax.set_title(
                title,
                fontsize=label_size,
            )

        plt.grid(True, color="gray", linestyle="dotted", which="both", alpha=0.5)
        plt.legend()

        outdir = get_date_dir(self.date)
        outpath = outdir / f"{savename}.pdf"

        plt.savefig(outpath, bbox_inches="tight")
        plt.close()

    def check_moon(self, coords):
        """
        Check proximity to the moon and moon illuminated fraction
        """
        moon = get_body("moon", self.midnight_utc, self.site)
        moon_sep = moon.separation(coords)
        illumination = apmoon.moon_illumination(self.midnight_utc)
        illumination_earlier = apmoon.moon_illumination(self.midnight_utc - (2 * u.day))
        illumination_later = apmoon.moon_illumination(self.midnight_utc + (2 * u.day))

        return {
            "sep": moon_sep.to(u.degree).value,
            "illum": illumination * 100,
            "illum-2": illumination_earlier,
            "illum+2": illumination_later,
        }

    @staticmethod
    def altitude_to_airmass(airmass: float | np.ndarray):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            altitude = 1.0 / np.cos(np.radians(90 - airmass))
        return altitude

    @staticmethod
    def airmass_to_altitude(altitude: float):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            airmass = 90 - np.degrees(np.arccos(1 / altitude))
        return airmass

    @staticmethod
    def get_moon_emoticon(moon_info: dict):
        illum = moon_info["illum"]
        illum_earlier = moon_info["illum-2"]
        illum_later = moon_info["illum+2"]
        if illum_later - illum_earlier > 0:
            waxing = True
        else:
            waxing = False
        if illum < 16.6:
            symbol = "🌑"
        elif illum < 33.3:
            if waxing:
                symbol = "🌒"
            else:
                symbol = "🌘"
        elif illum < 66.6:
            if waxing:
                symbol = "🌓"
            else:
                symbol = "🌗"
        elif illum < 83:
            if waxing:
                symbol = "🌔"
            else:
                symbol = "🌖"
        else:
            symbol = "🌕"

        return symbol



#!/usr/bin/env python3
# Author: Simeon Reusch (simeon.reusch@desy.de)
# License: BSD-3-Clause

import getpass
import logging
import os
import warnings
from os import environ

import keyring
from ztfquery import io  # type: ignore

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

try:
    keyring.get_password("dummy", f"dummyfi_user")
    # get_credentials("dummy", f"dummyfi_user")
    keyring_available = True
except keyring.errors.NoKeyringError:
    keyring_available = False

keyring_available = False


def get_credentials(service, token=False):
    credentials = {}
    if keyring_available is True:
        if token is False:
            username = keyring.get_password(service, f"{service}_user")
            password = keyring.get_password(service, f"{service}_password")

            if username is None:
                username = input(f"Enter your {service} login: ")

                password = getpass.getpass(
                    prompt=f"Enter your {service} password: ", stream=None
                )
                keyring.set_password(service, f"{service}_user", username)
                keyring.set_password(service, f"{service}_password", password)
                logger.info(f"Got {service} credentials")
            credentials.update({"user": username, "password": password})

        if token is True:
            token = keyring.get_password(service, f"{service}_token")
            if token is None:
                token = getpass.getpass(
                    prompt=f"Enter your {service} token: ", stream=None
                )
                keyring.set_password(service, f"{service}_token", token)
                logger.info(f"Got {service} credentials")
            credentials.update({"token": token})

    if keyring_available is False:
        if token is False:
            username = os.environ.get(f"{service}_user")
            password = os.environ.get(f"{service}_password")

            if username is None:
                username, password = io._load_id_(service)
            credentials.update({"user": username, "password": password})

        if token is True:
            token = os.environ.get(f"{service}_token")

            if token is None:
                logger.info(
                    f"This is a workaround using base64 obfuscation. If it asks for input: Enter an arbitrary username and the {service} token."
                )
                _, token = io._load_id_(service)
            credentials.update({"token": token})

    return credentials



#!/usr/bin/env python3
# Author: Simeon Reusch (simeon.reusch@desy.de)
# License: BSD-3-Clause

import re


def is_ztf_name(name: str) -> bool:
    """
    Checks if a string adheres to the ZTF naming scheme
    """
    if re.match(r"^ZTF[1-2]\d[a-z]{7}$", name):
        return True
    else:
        return False



#!/usr/bin/env python3
# Author: Simeon Reusch (simeon.reusch@desy.de)
# License: BSD-3-Clause

import getpass
import logging
import shutil
from pathlib import Path

import keyring
import requests

from betternot import fritz, io

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def get_finding_chart(ztf_id: str, date: str):
    """
    Download a finding chart from Fritz
    """

    date_full = date + "T12:00:00"
    url = f"/sources/{ztf_id}/finder?imsize=5&type=png&num_offset_stars=0&obstime={date_full}"

    logger.info(f"Issuing finding chart request for {ztf_id} and date {date}")
    response = fritz.api(method="get", url=url, stream=True)

    if response.status_code in (200, 400):
        outpath = io.get_date_dir(date)

        with open(outpath / f"{ztf_id}_{date.replace('-','_')}.png", "wb") as f:
            shutil.copyfileobj(response.raw, f)
        logger.info(
            f"Downloaded finding chart for {ztf_id} to {outpath / f'{ztf_id}_{date}.png'}"
        )
